/********************************************************
 * 					   - 3D-AMP -						*
 * 	  		3D Analysis and Measuring Plugin			*
 * 														*
 * 		Laetz, E.M.J., Rühr P.T., Bartolomaeus, T.,		*
 * 			Preisfeld, A. & Wägele, H. (2016)			*
 * 														*
 * 					   v. 1.2.0.						*
 *******************************************************/


if (isOpen("Log")) { 
     selectWindow("Log"); 
     run("Close"); 
} 
if (isOpen("Results")) { 
     selectWindow("Results"); 
     run("Close"); 
}

Stack.getDimensions(width, height, channels, slices, frames);
getPixelSize(orig_unit, orig_pixelWidth, orig_pixelHeight, orig_voxelDepth);

plugins = getDirectory("plugins");
unix = '/plugins/';
windows = '\\plugins\\';

if(endsWith(plugins, unix)){
	print("Running on Unix.");
	file = "Macros/3D-AMP_settings.txt";
}
if(endsWith(plugins, windows)){
	print("Running on Windows.");
	file = "Macros\\3D-AMP_settings.txt";
}

pathfile = plugins+file;

if(File.exists(pathfile) == true){
	filestring=File.openAsString(pathfile); 
	rows=split(filestring, "\n");
	r = 1;
	print("Loaded settings from " + pathfile + ".");
}
else{
	open_file = File.open(pathfile);
	print_array = newArray("3D-AMP Settings");
	for(i = 0; i < 100; i++){
		print(open_file, "0");
	}
	File.close(open_file)
	print("Created " + pathfile + "from default settings.");
}

filestring=File.openAsString(pathfile); 
rows=split(filestring, "\n");
r = 1;

Dialog.create("Presettings 1");
Dialog.addMessage("Welcome to 3D-AMP, the 3D Analysis and Measuring Plugin.");
Dialog.addMessage("Please choose your desired settings.");
Dialog.addMessage("___________________________________");
for (ch=1; ch<channels+1; ch++){
	Dialog.addCheckbox("Analyze channel" + ch, rows[r]);
	r++;
	Dialog.addNumber("Threshold", rows[r]);
	r++;
	Dialog.addCheckbox("Alignment", rows[r]);
	r++;
	Dialog.addCheckbox("Subtract signal from clouds; keep compact areas only", rows[r]);
	r++;
	Dialog.addMessage("___________________________________");
}
Dialog.addCheckbox("Define ROI", rows[r]);
r++;
Dialog.addCheckbox("Analyse areas inside AND outside ROI", rows[r]);
r++;
Dialog.addCheckbox("Show average", rows[r]);
r++;
Dialog.addCheckbox("Show explanation of Results", rows[r]);
r++;
Dialog.addCheckbox("Save results as Excel file", rows[r]);
r++;
Dialog.addCheckbox("Close image windows", rows[r]);
r++;
Dialog.addMessage("___________________________________");
Dialog.show();

check_channel = newArray("x");
thrsh_of_channel = newArray("x");
sig_subtr = newArray("x");
title_of_channel_wi = newArray("x");
title_of_channel_wo = newArray("x");
align_channel = newArray("x");

print_array = newArray("");
print_array[0] = "3D-AMP Settings";

for(i=0; i<channels; i++){
	curr_check_channel = Dialog.getCheckbox();
	check_channel = Array.concat(check_channel, curr_check_channel);
	print_array = Array.concat(print_array, curr_check_channel);

	curr_thrsh_of_channel = Dialog.getNumber();
	thrsh_of_channel = Array.concat(thrsh_of_channel, curr_thrsh_of_channel);
	print_array = Array.concat(print_array, curr_thrsh_of_channel);

	curr_align_channel = Dialog.getCheckbox();
	align_channel = Array.concat(align_channel, curr_align_channel);
	print_array = Array.concat(print_array, curr_align_channel);

	curr_sig_subtr = Dialog.getCheckbox();
	sig_subtr = Array.concat(sig_subtr, curr_sig_subtr);
	print_array = Array.concat(print_array, curr_sig_subtr);
}

user_ROI = Dialog.getCheckbox;
ins_out = Dialog.getCheckbox;
average = Dialog.getCheckbox();
show_explanation = Dialog.getCheckbox();
o_save = Dialog.getCheckbox();
o_close = Dialog.getCheckbox();

Dialog.create("Presettings 2");
Dialog.addString("Unit", rows[r]);

Dialog.addMessage("Current unit = "+orig_unit);
r++;
Dialog.addMessage("___________________________________");
Dialog.addNumber("Pixel size ", rows[r]);

Dialog.addMessage("Current pixel size = "+orig_pixelWidth);
r++;
Dialog.addMessage("___________________________________");
Dialog.addCheckbox("Save settings as default", rows[r]);
Dialog.addMessage("___________________________________");
Dialog.addMessage("Laetz, E.M.J., Ruehr P.T., Bartolomaeus, T., Preisfeld, A. & Waegele, H. (2016)")
Dialog.show();

unit = Dialog.getString();
px_size = Dialog.getNumber();
save_settings = Dialog.getCheckbox();

if(save_settings == true){
	print_array = Array.concat(print_array, user_ROI);
	print_array = Array.concat(print_array, ins_out);
	print_array = Array.concat(print_array, average);
	print_array = Array.concat(print_array, show_explanation);
	print_array = Array.concat(print_array, o_save);
	print_array = Array.concat(print_array, o_close);
	print_array = Array.concat(print_array, unit);
	print_array = Array.concat(print_array, px_size);
	print_array = Array.concat(print_array, save_settings);
	
	if(print_array.length < 99){
		for(c = print_array.length; c < 99; c++){
			print_array = Array.concat(print_array, "0");
		}
	}
	
	open_file = File.open(pathfile);
	for(i = 0; i < print_array.length; i++){
		if(print_array[i] != "x"){
			print(open_file, print_array[i]);
		}
	}
	File.close(open_file)
	print("Saved settings to " + pathfile + ".");
}

width_px = getWidth;
height_px = getHeight;

run("Properties...", "channels=channels slices=slices frames=frames unit=orig_unit pixel_width=px_size pixel_height=px_size voxel_depth=orig_voxelDepth");

if (user_ROI == true){
	waitForUser("1) Select an image from your stack and draw the Region of Interest (ROI)\ndirectly onto that image with any selection tool from the menue.\n2) Click 'Ok' AFTERWARDS."); 
}
else{
	run("Select All");
}	
run("Set Measurements...", "  area redirect=None decimal=3");
run("Measure");
area_ROI_um = getResult("Area", 0);
run("Clear Results");
if (isOpen("Results")) { 
     selectWindow("Results"); 
     run("Close"); 
}

img_res = width_px * height_px;
width_um = width_px * px_size;
height_um = height_px * px_size;
img_res_um = width_um * height_um;
perc_ROI_um = (area_ROI_um / img_res_um * 100);
area_ROI_px = (sqrt(area_ROI_um) / px_size) * (sqrt(area_ROI_um) / px_size);
area_wo_ROI_px = img_res - area_ROI_px;
area_wo_ROI_um = (sqrt(area_wo_ROI_px) * px_size) * (sqrt(area_wo_ROI_px) * px_size);

print_attributes(width_um, height_um, unit, width_px, height_px, px_size, area_ROI_um, area_ROI_px, area_wo_ROI_um, 
area_ROI_px, perc_ROI_um);
print_variables(channels, check_channel, thrsh_of_channel, align_channel, average, sig_subtr, o_close, o_save, unit);

getSelectionCoordinates(x, y);
run("Select None");

title_of_channel_wi = newArray("x");
title_of_channel_wo = newArray("x");

inputId_orig_stack = getImageID();
orig_stack = getTitle();

for(i=1; i<channels+1; i++){
	if(check_channel[i] == true){
		selectWindow(orig_stack);
		curr_duplicate = "duplicate channels=" + i;
		run("Duplicate...", curr_duplicate);
		title_of_channel_wi = Array.concat(title_of_channel_wi, getTitle());
		
		if(align_channel[i]){
			print("Alignment of channel " + i + " started...");
			run("StackReg", "transformation=Affine");
			run("Clear Results");
			print("-----------------");
		}

		if(sig_subtr[i] == true){
			run("Subtract Background...", "rolling=1 disable stack");
		}
		
		else{
			run("8-bit");
		}

		if(ins_out == true){
			run("Duplicate...", "duplicate");
			title_of_channel_wo = Array.concat(title_of_channel_wo, getTitle());
		}
	}
	else{
		title_of_channel_wi = Array.concat(title_of_channel_wi, "disabled");
		if(ins_out == true){
			title_of_channel_wo = Array.concat(title_of_channel_wo, "disabled");
		}
	}
}

print("Analysis started...");
for(i=1; i<channels+1; i++){
	if(check_channel[i] == true){
		selectWindow(title_of_channel_wi[i]);
		caption_1 = "Ch. " + i + " px > thrsh ins.ROI";
		caption_2 = "Ch. " + i + " A >thrsh ins.ROI";
		threshold = thrsh_of_channel[i];
		cl_inside = 0;
		cl_ROI(cl_inside, x, y);

		print("-----------------");
		print ("Starting analysis function for channel " + i + "...");

		analysis(width_px, height_px, slices, area_ROI_px, px_size, orig_stack, unit, average, caption_1, caption_2, 
threshold, cl_inside);

		selectWindow(title_of_channel_wi[i]);
		if(o_close){
			run("Close");
		}

		if(ins_out == true){
			selectWindow(title_of_channel_wo[i]);
			caption_1 = "Ch. " + i + " px > thrsh outs.ROI";
			caption_2 = "Ch. " + i + " A >thrsh outs.ROI";
			threshold = thrsh_of_channel[i];
			cl_inside = 1;
			cl_ROI(cl_inside, x, y);
			analysis(width_px, height_px, slices, area_ROI_px, px_size, orig_stack, unit, average, caption_1, 
caption_2, threshold, cl_inside);
		}
		
		print ("Analysis function for channel " + i + " done.");

		if(ins_out == true){
			selectWindow(title_of_channel_wo[i]);
			if(o_close){
				run("Close");
			}
		}
		
		setOption("ShowRowNumbers", false);
		updateResults;
	}
}
print("-----------------");

if (isOpen("Log")) { 
     selectWindow("Log");
}
if (isOpen("Results")) { 
     selectWindow("Results"); 
}

makeSelection("polygon", x, y);
beep();
print ("All done!");
print("-----------------");


if(o_save){
	Dialog.create("Filename");
	Dialog.addString("Please enter name for results table file", orig_stack);
	Dialog.show();
	output_name = Dialog.getString;
	directory = getDirectory("Select output directory for results table.");
	path = directory+output_name;
	output_suffix = ".xls";
	output_file = path+output_suffix;
	saveAs("Results", output_file);
	print("Saved Results to");
	print(output_file);
}

if(show_explanation == true){
	Dialog.create("Results table");
	Dialog.addMessage(" Explanation of the Results table:");
	Dialog.addMessage("");
	Dialog.addMessage("The unit depends on the unit (here: " + unit + ") you defined before.");
	Dialog.addMessage("******************************************************************************");
	Dialog.addMessage("Image name                     = name of the image");
	Dialog.addMessage("******************************************************************************");
	Dialog.addMessage("px.>thrsh [%] ins.ROI          = percentage of pixels that are above the threshold inside the ROI");
	Dialog.addMessage("A.>thrsh. [um^2] ins.ROI     = area covered by pixels above thresohld within ROI");
	if(ins_out == true){
		Dialog.addMessage("px.>thrsh. [%] outs.ROI        = percentage of pixels that are above the threshold outside the ROI");
		Dialog.addMessage("A.>thrsh. [um^2] outs.ROI    = area covered by pixels above threshold outside ROI");
	}
	Dialog.addMessage("******************************************************************************");
	Dialog.show;
};


Dialog.create("Rerun or end");
Dialog.addChoice("What do you want to do now?", newArray("Rerun 3D-AMP", "End"));
Dialog.show();
next = Dialog.getChoice();

if(next == 'Rerun 3D-AMP'){
	waitForUser("Wait for user", "If not done already, please save the results table!\n!!! Clicking \"OK\" will close the results and restart 3D AMP !!!");
	run("3D-AMP 1-2-0");
}

/**************************************************** - FUNCTIONS - ****************************************************/

function cl_ROI(cl_inside, x, y){
	makeSelection("polygon", x, y);
	setBackgroundColor(0, 0, 0);
	if(cl_inside){
		run("Clear", "stack");
	}
	else {
		run("Clear Outside", "stack");
	}
}

function analysis(width_px, height_px, slices, area_ROI_px, px_size, orig_stack, unit, average, caption_1, caption_2, 
threshold, cl_inside){
	count_px = 0;
	acc_perc_ab_thrsh = 0;
	acc_area_um_ab_thrsh = 0;
	count_img = 1;
	for (i=1; i<frames+1; i++){
		Stack.setFrame(i); 
		for (j=1; j<slices+1; j++) { 
			Stack.setSlice(j); 
			for (x = 0; x < width_px; x++) { 
				for(y = 0; y < height_px; y++) { 
					curr_px = getPixel(x, y); 
					if (curr_px > threshold) 
						count_px++;
	        	}
	    	}
	    	
	    if(cl_inside){
	    	perc_ab_thrsh = count_px / area_wo_ROI_px * 100;
	    }
	    else{
			perc_ab_thrsh = count_px / area_ROI_px * 100;
	    }

		area_um_ab_thrsh = count_px * px_size * px_size;
		acc_perc_ab_thrsh = acc_perc_ab_thrsh + perc_ab_thrsh;
	    acc_area_um_ab_thrsh = acc_area_um_ab_thrsh + area_um_ab_thrsh;
		table_img_no = count_img - 1;
		curr_title_no = orig_stack+"_"+count_img;
		setResult("Image name", table_img_no, curr_title_no);
		setResult(caption_1 + " [" + "%]", table_img_no, perc_ab_thrsh);
		setResult(caption_2 + " [" + unit + "^2]", table_img_no, area_um_ab_thrsh);
		updateResults;
		count_px = 0;
		count_img++;
		}	
	}
	
	if (average){
	acc_perc_ab_thrsh = acc_perc_ab_thrsh / (table_img_no + 1);
	setResult("Image name", (count_img - 1), "mean [%] or sum [ " + unit + "^2]");
	setResult(caption_1 + " [" + "%]", (count_img - 1), acc_perc_ab_thrsh);
	setResult(caption_2 + " [" + unit + "^2]", (count_img - 1), acc_area_um_ab_thrsh);
	}
}

function print_attributes(width_um, height_um, unit, width_px, height_px, px_size, area_ROI_um, area_ROI_px, area_wo_ROI_um, area_ROI_px, perc_ROI_um){
	print("**************************************************************************************");
	print("- 3D-AMP - 3D Analysis and Measuring Plugin - ");
	print("**************************************************************************************");
	print("**************************************************************************************");
	print("Image resolution: " + width_um + " * " + height_um + " " + unit + " = " + width_px + " * " + height_px + " px");
	print("   --> "+ px_size + " " + unit + " / px");
	print("ROI = " + area_ROI_um + " " + unit + "^2 = " + area_ROI_px + " px^2");
	print("Area outs. ROI = " + area_wo_ROI_um + " " + unit + "^2 = " + area_wo_ROI_px + " px^2");
	print("ROI/img resolution [%] = " + perc_ROI_um);
	print("**************************************************************************************");
}

function print_variables(channels, check_channel, thrsh_of_channel, align_channel, average, sig_subtr, o_close, o_save, unit){
	print("-----------------");
	for(i=1; i<channels+1; i++){
		if(check_channel[i] == true){
			print("Channel " + i);
			print("Threshold: " + thrsh_of_channel[i]);
			print("Alignment:\t\t\t" + align_channel[i]);
			print("Background subtraction:\t" + sig_subtr[i]);
			print("-----------------");
		}
		else{
			print("Channel " + i + ": disabled.");
			print("-----------------");
		}
	}
	print("Image width: " + width_um + " " + unit);
	print("Display mean values in results:\t" + average);
	print("Save results:\t" + o_save);
	print("Show explanation of results:\t" + show_explanation);
	print("Close windows:\t" + o_close);
	print("**************************************************************************************");	
}